@@ -14,6 +14,7 @@ import argparse
 import collections
 import cProfile as profile
 import copy
+import dataclasses
 import difflib
 import fnmatch
 import functools
@@ -29,9 +30,9 @@ import multiprocessing
 import operator
 import os
 import os.path as osp
+import time
 import types
 from os.path import abspath, basename, dirname, expanduser, exists, isfile, join, splitext
-# from pprint import pprint, pformat
 import platform
 import plistlib
 import pprint as pp
@@ -124,6 +125,37 @@ class BandPassLogFilter(object):
         return self.__levelbounds[0] <= log.levelno <= self.__levelbounds[1]
 
 
+def get_platform_home_dir():
+    plat = platform.system()
+    if plat == 'Windows':
+        return os.getenv('USERPROFILE')
+    elif plat == 'Darwin':
+        return osp.expanduser('~/')
+    elif plat == 'Linux':
+        return osp.expanduser('~/')
+    raise NotImplementedError(f'unsupported platform: {plat}')
+
+
+def get_platform_appdata_dir(winroam=True):
+    plat = platform.system()
+    if plat == 'Windows':
+        return os.getenv('APPDATA' if winroam else 'LOCALAPPDATA')
+    elif plat == 'Darwin':
+        return osp.expanduser('~/Library/Application Support')
+    raise NotImplementedError(f'unsupported platform: {plat}')
+
+
+def get_platform_tmp_dir():
+    plat = platform.system()
+    if plat == 'Windows':
+        return join(os.getenv('LOCALAPPDATA'), 'Temp')
+    elif plat == 'Darwin':
+        return join(expanduser('~'), 'Library', 'Caches')
+    elif plat == 'Linux':
+        return '/tmp'
+    raise NotImplementedError(f'unsupported platform: {plat}')
+
+
 def build_default_logger(logdir, name=None, cfgfile=None, verbose=False):
     """
     Create per-file logger and output to shared log file.
@@ -136,7 +168,7 @@ def build_default_logger(logdir, name=None, cfgfile=None, verbose=False):
     :return: logger object.
     """
     os.makedirs(logdir, exist_ok=True)
-    cfg_file = cfgfile or join(_script_dir, 'logging.json')
+    cfg_file = cfgfile or osp.join(_script_dir, 'logging.json')
     try:
         if sys.version_info.major > 2:
             with open(cfg_file, 'r', encoding=TXT_CODEC,
@@ -147,10 +179,10 @@ def build_default_logger(logdir, name=None, cfgfile=None, verbose=False):
                 text = f.read()
         # Add object_pairs_hook=coll.OrderedDict hook for py3.5 and lower.
         logging_config = json.loads(text, object_pairs_hook=collections.OrderedDict)
-        logging_config['handlers']['file']['filename'] = join(logdir, logging_config['handlers']['file']['filename'])
+        logging_config['handlers']['file']['filename'] = osp.join(logdir, logging_config['handlers']['file']['filename'])
     except Exception:
         filename = name or basename(basename(logdir.strip('\\/')))
-        log_path = join(logdir, '{}.log'.format(filename))
+        log_path = osp.join(logdir, '{}.log'.format(filename))
         logging_config = {
             "version": 1,
             "disable_existing_loggers": False,
@@ -218,13 +250,13 @@ def build_default_logger(logdir, name=None, cfgfile=None, verbose=False):
     return logging.getLogger(name or 'default')
 
 
-_logger = build_default_logger(logdir=join(_script_dir, os.pardir, 'temp'), name=splitext(basename(__file__))[0])
+glogger = build_default_logger(logdir=osp.join(get_platform_tmp_dir(), '_util'), name='util', cfgfile=None, verbose=False)
 
 
 def catch_unknown_exception(exc_type, exc_value, exc_traceback):
     """Global exception to handle uncaught exceptions"""
     exc_info = exc_type, exc_value, exc_traceback
-    _logger.error('Unhandled exception: ', exc_info=exc_info)
+    glogger.error('Unhandled exception: ', exc_info=exc_info)
     # _logger.exception('Unhandled exception: ')  # try-except block only.
     # sys.__excepthook__(*exc_info)  # Keep commented out to avoid msg dup.
 
@@ -521,7 +553,7 @@ def get_md5_checksum(file):
     return myhash.hexdigest()
 
 
-def logprogress(msg='Task', loghook=_logger.info, errorhook=_logger.error):
+def logprogress(msg='Task', loghook=glogger.info, errorhook=glogger.error):
     def wrap(function):
         @functools.wraps(function)
         def wrapper(*args, **kwargs):
@@ -537,7 +569,7 @@ def logprogress(msg='Task', loghook=_logger.info, errorhook=_logger.error):
     return wrap
 
 
-def logcall(msg='trace', logger=_logger):
+def logcall(msg='trace', logger=glogger):
     def wrap(function):
         @functools.wraps(function)
         def wrapper(*args, **kwargs):
@@ -614,16 +646,16 @@ def execute_concurrency(worker, shared, lock, algorithm):
         - Args: worker input args
         - Result: worker returned results in order
     """
-    global _logger
+    global glogger
     # TODO: measure timeout for .join()
     if algorithm['Type'] == 'Sequential':
         results = []
         for t, task in enumerate(shared['Tasks']):
-            _logger.debug('Execute {} in order: {} of {}: {}'.format(shared['Title'], t+1, len(shared['Tasks']), task['Title']))
+            glogger.debug('Execute {} in order: {} of {}: {}'.format(shared['Title'], t + 1, len(shared['Tasks']), task['Title']))
             results.append(worker(task))
         return [result[1] for result in results]
     elif algorithm['Type'] == 'Process':
-        _logger.debug('Execute {} in pool of {} processes ...'.format(shared['Title'], algorithm['Count']))
+        glogger.debug('Execute {} in pool of {} processes ...'.format(shared['Title'], algorithm['Count']))
         #
         # Known Issue:
         # - https://bugs.python.org/issue9400
@@ -643,7 +675,7 @@ def execute_concurrency(worker, shared, lock, algorithm):
         return [result[1] for result in results]
     elif algorithm['Type'] == 'Thread':
         import concurrent.futures
-        _logger.debug('Execute {} in pool of {} threads ...'.format(shared['Title'], algorithm['Count']))
+        glogger.debug('Execute {} in pool of {} threads ...'.format(shared['Title'], algorithm['Count']))
         executor = concurrent.futures.ThreadPoolExecutor(max_workers=algorithm['Count'])
         results = executor.map(worker, shared['Tasks'])
         return [result[1] for result in results]
@@ -669,35 +701,7 @@ def profile_runs(funcname, modulefile, nruns=5):
     stats.print_stats()
 
 
-def get_platform_home_dir():
-    plat = platform.system()
-    if plat == 'Windows':
-        return os.getenv('USERPROFILE')
-    elif plat == 'Darwin':
-        return osp.expanduser('~/')
-    elif plat == 'Linux':
-        return osp.expanduser('~/')
-    raise NotImplementedError(f'unsupported platform: {plat}')
-
-
-def get_platform_appdata_dir(winroam=True):
-    plat = platform.system()
-    if plat == 'Windows':
-        return os.getenv('APPDATA' if winroam else 'LOCALAPPDATA')
-    elif plat == 'Darwin':
-        return osp.expanduser('~/Library/Application Support')
-    raise NotImplementedError(f'unsupported platform: {plat}')
-
-
-def get_platform_tmp_dir():
-    plat = platform.system()
-    if plat == 'Windows':
-        return join(os.getenv('LOCALAPPDATA'), 'Temp')
-    elif plat == 'Darwin':
-        return join(expanduser('~'), 'Library', 'Caches')
-    elif plat == 'Linux':
-        return '/tmp'
-    raise NotImplementedError(f'unsupported platform: {plat}')
+glogger = build_default_logger(osp.join(get_platform_tmp_dir(), '_util'), name='util', cfgfile=None, verbose=False)
 
 
 def write_plist_fields(cfg_file, my_map):
@@ -755,12 +759,10 @@ def convert_to_wine_path(path, drive=None):
     full_path = abspath(expanduser(path))
     home_folder = os.environ['HOME']
     if leading_homefolder := full_path.startswith(home_folder):
-        mapped_drive = 'Y:'
+        mapped_drive = drive or 'Y:'
         full_path = full_path.removeprefix(home_folder)
     else:
-        mapped_drive = 'Z:'
-    if drive:
-        mapped_drive = drive
+        mapped_drive = drive or 'Z:'
     full_path = full_path.replace('/', '\\')
     return mapped_drive + full_path
 
@@ -814,7 +816,7 @@ def match_files_except_lines(file1, file2, excluded=None):
 
 class RerunLock:
     """Lock process from reentering when seeing lock file on disk."""
-    def __init__(self, name, folder=None, logger=_logger):
+    def __init__(self, name, folder=None, logger=glogger):
         os.makedirs(folder, exist_ok=True)
         filename = f'lock_{name}.json' if name else 'lock_{}.json'.format(next(tempfile._get_candidate_names()))
         self.lockFile = osp.join(folder, filename) if folder else join(get_platform_tmp_dir(), filename)
@@ -880,7 +882,7 @@ class RerunLock:
         raise RuntimeError(msg)
 
 
-def rerun_lock(name, folder=None, logger=_logger):
+def rerun_lock(name, folder=None, logger=glogger):
     """Decorator for reentrance locking on functions"""
     def decorator(f):
         @functools.wraps(f)
@@ -999,7 +1001,7 @@ def remove_from_os_paths(bindir, usesyspath=True, inmemonly=False):
     os.environ[path_var] = os.environ[path_var].replace(bindir, '')
 
 
-def run_cmd(cmd, cwd=None, logger=None, check=True, shell=False):
+def run_cmd(cmd, cwd=None, logger=None, check=True, shell=False, verbose=False):
     """
     Use shell==True with autotools where new shell is needed to treat the entire command option sequence as a command,
     e.g., shell=True means running sh -c ./configure CFLAGS="..."
@@ -1007,19 +1009,26 @@ def run_cmd(cmd, cwd=None, logger=None, check=True, shell=False):
     local_debug = logger.debug if logger else print
     local_info = logger.info if logger else print
     local_error = logger.error if logger else print
-    local_debug(f"""\
+    console_info = local_info if logger and verbose else local_debug
+    # show cmdline with or without exceptions
+    cmd_log = f"""\
 {' '.join(cmd)}
 cwd: {osp.abspath(cwd) if cwd else os.getcwd()}
-""")
+"""
+    local_info(cmd_log)
     try:
         proc = subprocess.run(cmd, check=check, shell=shell, stdout=subprocess.PIPE, stderr=subprocess.PIPE, cwd=cwd)
-        log = proc.stdout.decode(TXT_CODEC)
-        local_debug(log)
+        stdout_log = proc.stdout.decode(TXT_CODEC)
+        stderr_log = proc.stderr.decode(TXT_CODEC)
+        if stdout_log:
+            console_info(f'stdout:\n{stdout_log}')
+        if stderr_log:
+            local_error(f'stderr:\n{stderr_log}')
     except subprocess.CalledProcessError as e:
-        log = f'stdout: {e.stdout.decode(TXT_CODEC)}'
-        err = f'stderr: {e.stderr.decode(TXT_CODEC)}'
-        local_info(log)
-        local_error(err)
+        stdout_log = f'stdout:\n{e.stdout.decode(TXT_CODEC)}'
+        stderr_log = f'stderr:\n{e.stderr.decode(TXT_CODEC)}'
+        local_info(stdout_log)
+        local_error(stderr_log)
         raise e
     except Exception as e:
         # no need to have header, exception has it all
@@ -1476,7 +1485,7 @@ def copy_file(src, dst, isdstdir=False, keepmeta=False):
     try:
         copyfunc(src, dst)
     except shutil.SameFileError:
-        _logger.warning(f'source and destination are identical. will SKIP: {osp.abspath(src)} -> {osp.abspath(dst)}.')
+        glogger.warning(f'source and destination are identical. will SKIP: {osp.abspath(src)} -> {osp.abspath(dst)}.')
 
 
 def move_file(src, dst, isdstdir=False):
@@ -1488,7 +1497,7 @@ def move_file(src, dst, isdstdir=False):
     try:
         shutil.move(src, dst)
     except shutil.SameFileError:
-        _logger.warning(f'source and destination are identical. will SKIP: {osp.abspath(src)} -> {osp.abspath(dst)}.')
+        glogger.warning(f'source and destination are identical. will SKIP: {osp.abspath(src)} -> {osp.abspath(dst)}.')
 
 
 def compare_dirs(dir1, dir2, ignoreddirpatterns=(), ignoredfilepatterns=(), showdiff=True):
@@ -1812,7 +1821,13 @@ class Autotools:
         run_cmd(['make', 'clean'], cwd=self.pkgRoot, logger=self.logger)
 
 
-def build_apple_iconset(master, iconset):
+def validate_platform(plat):
+    if platform.system() != plat:
+        raise NotImplementedError(f'Runs on {plat} only.')
+
+
+def build_iconset(master, iconset):
+    validate_platform('Darwin')
     iconset_dir = osp.join(f'/{get_platform_tmp_dir()}/icons/icon.iconset')
     os.makedirs(iconset_dir, exist_ok=True)
     sizes = ('16', '16@2x', '32', '32@2x', '128', '128@2x', '256', '256@2x', '512')
@@ -1827,6 +1842,241 @@ def build_apple_iconset(master, iconset):
     shutil.rmtree(iconset_dir, ignore_errors=True)
 
 
+def fix_dylib_dependencies(target, rootprefix=''):
+    """
+    prepare .dylibs for macOS app deployment
+    - pre-condition: all the fixing must happen at the dst locations, i.e., requiring pre-deployment
+    - this is because the fixing requires src libs to exist and maintains old embedded paths
+    - recursively find all dependencies of a target binary executable or lib
+    - replace build-time prefix with user-supplied string for all dependencies
+    - replace dylib's own id with user-provided prefix
+    """
+    class DynLinked:
+        """
+        a pre-deployed binary file having dependencies to fix for runtime execution
+        """
+        analyzer = None
+        fixer = None
+        execPath = None  # physical runtime exec folder
+        syslibPrefixes = ['/System/Library', '/usr/']
+
+        def __init__(self, buildproduct, prefix='@executable_path', parent=None):
+            assert osp.isabs(buildproduct), f'Not an absolute path: {buildproduct}, parent: {parent}'
+            self.buildProduct = buildproduct
+            self.distributable = None
+            self.deps = None
+            # CAUTION
+            # - prefix is to be embedded into binaries
+            # - input prefix must be relative to @executable_path, e.g., @executable_path/libs
+            # - input may already start with a build-time base
+            runtime_base_tag = '@executable_path'
+            # prefix must start with '@executable_path'
+            self.prefix = prefix.rstrip(os.sep) if prefix.startswith(runtime_base_tag) else osp.join(runtime_base_tag, prefix.strip(os.sep))
+            # for reference only, we trace deps tree top-down from the outside
+            self.ref = parent
+            if is_root := not self.ref:
+                # root binary is our target, no need to decorate
+                self.distributable = buildproduct
+                DynLinked.execPath = osp.dirname(self.distributable)
+                glogger.info(f'Root aka. target binary: {self.distributable}')
+            else:
+                assert osp.isdir(DynLinked.execPath), 'Root binary has not been processed'
+                # this path may not exist yet
+                # @executable_path/relative/to/my.dylib
+                dest_dir = osp.join(DynLinked.execPath, relpath_to_exepath := osp.relpath(prefix, runtime_base_tag).strip(os.sep))
+                self.distributable = osp.join(dest_dir, bin_filename := osp.split(self.buildProduct)[1])
+
+        def extract_direct_deps(self):
+            """
+            - get 1st-order dependencies using otool -L
+            """
+            # all deps of a leaf binary are system libs and so their deps are empty
+            if sys_lib := not osp.isfile(self.buildProduct):
+                return []
+            # CAUTION:
+            # - install_name_tool requires source binaries to exist at their embedded src paths
+            # - so we must copy all the src bins to prefix location (dst) first to preserve embedded src paths
+            if not osp.isfile(self.distributable):
+                copy_file(self.buildProduct, self.distributable, isdstdir=False)
+            proc = run_cmd([DynLinked.analyzer, '-L', self.distributable])
+            list_head_lineno = 1
+            deps = [line.strip() for line in self._extract_deps(proc.stdout)[list_head_lineno:]]
+            # remove suffixes from dependency line
+            deps = list(map(lambda d: d[:d.find(' (compatibility version')], deps))
+            deps = list(map(lambda d: d[:start] if (start := d.find(' (architecture')) >= 0 else d, deps))
+            # bypass:
+            # - sys deps: they are not shipped and should be ignored
+            # - self: it requires a different cmd to fix, dylib only
+            to_bypass = list(filter(lambda d: any([d.startswith(path) for path in DynLinked.syslibPrefixes + [self.buildProduct]]), deps))
+            deps = list(set(deps).difference(to_bypass))
+            if not self.deps:
+                self.deps = deps
+            else:
+                glogger.info('Deps exist. Skipped to avoid pollution from already-fixed @executable_path prefix')
+            glogger.info(f"""Dependencies found for {self.buildProduct}:
+System and self (bypassed):
+{pp.pformat(to_bypass, indent=2)}
+Application (fixable) :
+{pp.pformat(deps, indent=2)}""")
+            return deps  # always return latest analytical result
+
+        def fix_deps(self):
+            """
+            recursively fix deps' paths with global prefix:
+            - install-dependent dylib search path
+            """
+            assert osp.isfile(self.distributable)
+            for dep in self.deps:
+                assert dep.endswith('.dylib')
+                distributable = osp.join(self.prefix, osp.split(dep)[1])
+                run_cmd([DynLinked.fixer, '-change', dep, distributable, dep_ref := self.distributable])
+
+        def fix_self(self):
+            """
+            - The first line of otool output is the binary itself.
+            - It requires a unique way of fixing
+            """
+            # copy src libs to dst in advance to avoid polluting src libs while iterating
+            # - translate prefix if it starts with @executable_path
+            assert osp.isfile(self.distributable)
+            distributable = osp.join(self.prefix, fn := osp.split(self.buildProduct)[1])
+            # rename parent dir
+            run_cmd([DynLinked.fixer, '-id', distributable, self.distributable])
+
+        def report(self):
+            if sys_lib := not osp.isfile(self.distributable):
+                return
+            proc = run_cmd([DynLinked.analyzer, '-L', self.distributable])
+            glogger.info(f'fixed for {self.distributable}:\n{proc.stdout.decode(TXT_CODEC)}')
+
+        def _extract_deps(self, otoolout):
+            return otoolout.decode(TXT_CODEC).split('\n\t')
+
+    def _collect_deps_tree(my_bin, io_allbins, prefix):
+        """
+        recursively find input binary's dependencies
+        :param my_bin: input binary as the root
+        :param io_allbins: dict to save all binaries and their deps for fixing them next
+        :return:
+        """
+        if not isinstance(my_bin, DynLinked):
+            raise TypeError('Input binary must be a DynLinked type')
+        if exit_at_leaf := not my_bin.extract_direct_deps():
+            return False
+        parent = copy.deepcopy(my_bin)
+        io_allbins[parent.buildProduct] = parent
+        for child in parent.deps:
+            if cyclic_refs := child == parent.buildProduct:
+                continue
+            if already_fixed := child.startswith(prefix) or child.startswith('@executable_path'):
+                glogger.debug(f'Already fixed by other exec: {child}, ref: {parent.buildProduct}; Skipped')
+                continue
+            glogger.debug(f'Found child: {child} of {parent.buildProduct}')
+            cbin = DynLinked(child, prefix=prefix, parent=parent)
+            _collect_deps_tree(cbin, io_allbins, prefix)
+        return True
+    validate_platform('Darwin')
+    if not osp.isfile(target):
+        raise FileNotFoundError(f'Missing target: {target}')
+    target = osp.abspath(target)
+    if osp.isabs(rootprefix):
+        raise ValueError(f'Expected prefix to be relative path or @executable_path, got absolute: {rootprefix}')
+    DynLinked.fixer = shutil.which('install_name_tool')
+    DynLinked.analyzer = shutil.which('otool')
+    if not DynLinked.fixer or not DynLinked.analyzer:
+        raise FileNotFoundError(f'Missing install_name_tool or otool. Install Xcode and toolchain first')
+    all_bins = {}
+    root = DynLinked(target, prefix=rootprefix)
+    if _collect_deps_tree(root, all_bins, prefix=rootprefix):
+        glogger.info('dependency recursively parsed')
+    for key in all_bins:
+        all_bins[key].fix_deps()
+    # CAUTION: must fix all deps before fixing self
+    for key in all_bins:
+        all_bins[key].fix_self()
+    for key in all_bins:
+        all_bins[key].report()
+
+
+def codesign(binary, identity=None, overwrite=True):
+    """
+    identity looks like: e.g., "Apple Development: me@email.com (IDIDIDIDID)"
+    """
+    validate_platform('Darwin')
+    run_cmd(['security', 'find-identity', '-v', '-p', 'codesigning'])
+    if overwrite:
+        run_cmd(['codesign', '--remove-signature', binary])
+    # -v with feedback
+    run_cmd(['codesign', '-s', identity, '-v', binary])
+    # gatekeeper validation
+    run_cmd(['spctl', '-a', '-t', 'exec', '-vv', binary])
+
+
+def build_dmg(masterdir, resdir='', dmg='', name=''):
+    """
+    build a DMG installer based on fixed layout, and lazy-include resources
+    - masterdir: content folder holding the app and optionally sub-folder "help"
+    - resdir: asset folder for building dmg only, not part of app content, usually holds background image (bg.png, made by Keynote 4:3, size 1024x768) and volume iconset (app.icns)
+    - dmg: path to output dmg
+    - name: app name, default to use basename of masterdir
+    """
+    validate_platform('Darwin')
+    app = types.SimpleNamespace(name='', helpDir='', license='', bundle='')
+    res = types.SimpleNamespace(background='', icon='', mountVolume='')
+    app.name = name or osp.basename(masterdir)
+    app.bundle = f'{app.name}.app'
+    app.helpDir = osp.join(masterdir, 'help')
+    if osp.isdir(app.helpDir):
+        app.license = osp.join(app.helpDir, 'LICENSE.txt')
+    res.mountVolume = f'{app.name}_installer'
+    if osp.isdir(resdir):
+        res.background = osp.join(resdir, 'bg.png')
+        res.icon = osp.join(resdir, 'app.icns')
+        assert osp.isfile(res.background)
+        assert osp.isfile(res.icon)
+    out_dmg = dmg or osp.join(osp.dirname(masterdir), f'install_{app.name}.dmg'.lower())
+    dmg_tool = shutil.which('create-dmg')
+    if not dmg_tool:
+        raise FileNotFoundError('Missing create-dmg; install and retry')
+    cmd = [dmg_tool,
+           '--volname', res.mountVolume,
+           '--volicon', res.icon,
+           '--icon-size', '90',
+           '--hide-extension', app.bundle, ]
+    if osp.isfile(app.license):
+        cmd += ['--eula', app.license]
+    window_rect = types.SimpleNamespace(x='100', y='0', w='500', h='300')
+    icon_pos = types.SimpleNamespace(x='150', y='110')
+    appdroplink_pos = types.SimpleNamespace(x='350', y='110')
+    help_pos = types.SimpleNamespace(x='', y='')
+    if osp.isdir(app.helpDir):
+        window_rect.w, window_rect.h = '550', '500'
+        icon_pos.x, icon_pos.y = '150', '150'
+        help_pos.x, help_pos.y = '150', '300'
+        appdroplink_pos.x, appdroplink_pos.y = '400', '150'
+    if osp.isfile(res.background):
+        glogger.warning('Found background image; size must be 1024x768 (Keynote 4:3)')
+        cmd += ['--background', res.background]
+        window_rect.w, window_rect.h = '1024', '820'
+        icon_pos.x, icon_pos.y = '362', '378'
+        help_pos.x, help_pos.y = '362', '500'
+        appdroplink_pos.x, appdroplink_pos.y = '650', '378'
+    # must add help after adjusting for background
+    if osp.isdir(app.helpDir):
+        cmd += ['--add-file', 'Help', app.helpDir, help_pos.x, help_pos.y]
+    cmd += ['--window-pos', window_rect.x, window_rect.y,
+            '--window-size', window_rect.w, window_rect.h,
+            '--icon', app.bundle, icon_pos.x, icon_pos.y,
+            '--app-drop-link', appdroplink_pos.x, appdroplink_pos.y, ]
+    cmd += [out_dmg, masterdir]
+    if tmp_dmgs := glob.glob('rw.*.dmg'):
+        for td in tmp_dmgs:
+            os.remove(td)
+    if osp.isfile(out_dmg):
+        os.remove(out_dmg)
+    run_cmd(cmd)
+
+
 def _test():
     pass
 
